
from .GameMaster import GameMaster


class Move:

    def __init__(self, *args, pvp=False, **kwargs):
        '''
        The named move paramters can be passed via keyword arguments,
        or via a str or dict as the first positional argument:

            Move("Counter"), Move({"name": "Counter"}), Move(name="Counter")

        give the same thing.
        '''
        game_master = GameMaster.CurrentInstance
        if len(args) > 0:
            if isinstance(args[0], str):
                kwargs["name"] = args[0]
                args = args[1:]
            elif isinstance(args[0], dict):
                kwargs.update(args[0])
                args = args[1:]
        if "name" in kwargs:
            if pvp:
                kwargs.update(game_master.search_pvp_move(kwargs["name"]))
            else:
                kwargs.update(game_master.search_pve_move(kwargs["name"]))
        self.pokeType = kwargs["pokeType"]
        self.power = kwargs["power"]
        self.energy = kwargs["energy"]
        self.duration = kwargs["duration"]
        self.dws = kwargs["dws"]
        if "effect" in kwargs:
            self.effect = kwargs["effect"]




from .GameMaster import GameMaster
from .Move import Move

ROLE_PVE_ATTACKER = "ae"
ROLE_PVP_ATTACKER = "ap"
ROLE_GYM_DEFENDER = "gd"
ROLE_RAID_BOSS = "rb"


class Pokemon:

    @staticmethod
    def calc_cp(bAtk, bDef, bStm, cpm, atkiv, defiv, stmiv):
        Atk = (bAtk + atkiv) * cpm
        Def = (bDef + defiv) * cpm
        Stm = (bStm + stmiv) * cpm
        return max(10, int(Atk * (Def * Stm)**0.5 / 10))

    @staticmethod
    def infer_cpm_and_IVs(bAtk, bDef, bStm, target_cp):
        CPMultipliers = GameMaster.CurrentInstance.CPMultipliers
        closest = None
        closest_cp = 0
        min_cpm_i = 0
        max_cpm_i = len(CPMultipliers) - 1
        while min_cpm_i < max_cpm_i:
            if Pokemon.calc_cp(bAtk, bDef, bStm, CPMultipliers[min_cpm_i], 15, 15, 15) >= target_cp:
                break
            min_cpm_i += 1
        while max_cpm_i > min_cpm_i:
            if Pokemon.calc_cp(bAtk, bDef, bStm, CPMultipliers[max_cpm_i], 0, 0, 0) <= target_cp:
                break
            max_cpm_i -= 1
        for cpm in CPMultipliers[min_cpm_i: max_cpm_i + 1]:
            for stmiv in range(16):
                for defiv in range(16):
                    for atkiv in range(16):
                        cp = Pokemon.calc_cp(
                            bAtk, bDef, bStm, cpm, atkiv, defiv, stmiv)
                        if cp == target_cp:
                            return (cpm, atkiv, defiv, stmiv)
                        elif closest_cp < cp and cp < target_cp:
                            closest_cp = cp
                            closest = (cpm, atkiv, defiv, stmiv)
        return closest

    def __init__(self, *args, **kwargs):
        '''
        All named parameters should be passed via keyword arguments, including:
            name, fmove, cmove, cmoves, level, atkiv, defiv, stmiv, cp, role, tier, immortal, strategy

        Some examples:
            Define an attacker:
                (name, fmove, cmove/cmoves, level, atkiv, defiv, stmiv)
            Define an attacker by cp (infer the stats):
                (name, fmove, cmove/cmoves, cp)
            Define a raid boss:
                (name, fmove, cmove, role=ROLE_RAID_BOSS, tier)
            Define a gym defender by cp (infer the stats, too):
                (name, fmove, cmove, cp, role=ROLE_GYM_DEFENDER)
        '''
        game_master = GameMaster.CurrentInstance
        if len(args) > 0:
            if isinstance(args[0], str):
                kwargs["name"] = args[0]
                args = args[1:]
            elif isinstance(args[0], dict):
                kwargs.update(args[0])
                args = args[1:]
        if "name" in kwargs:
            kwargs.update(game_master.search_pokemon(kwargs["name"]))

        role = kwargs.get("role", ROLE_PVE_ATTACKER)
        self.tier = None
        if role == ROLE_RAID_BOSS or "tier" in kwargs:
            self.tier = str(kwargs["tier"])
            tier_setting = game_master.search_raid_tier(self.tier)
            kwargs["attack"] = (kwargs["baseAtk"] + 15) * tier_setting["cpm"]
            kwargs["defense"] = (kwargs["baseDef"] + 15) * tier_setting["cpm"]
            kwargs["maxHP"] = tier_setting["maxHP"]
        elif len(args) < 5:
            if "cp" in kwargs:
                cp = kwargs["cp"]
                if not isinstance(cp, int):
                    raise TypeError("Expected int, got {}".format(type(cp)))
                cpm, atkiv, defiv, stmiv = Pokemon.infer_cpm_and_IVs(
                    kwargs["baseAtk"], kwargs["baseDef"], kwargs["baseStm"], cp)
            else:
                cpm = game_master.search_cpm(kwargs.get("level", 40))
                atkiv = int(kwargs.get("atkiv", 15))
                defiv = int(kwargs.get("defiv", 15))
                stmiv = int(kwargs.get("stmiv", 15))
            kwargs["attack"] = (kwargs["baseAtk"] + atkiv) * cpm
            kwargs["defense"] = (kwargs["baseDef"] + defiv) * cpm
            kwargs["maxHP"] = int((kwargs["baseStm"] + stmiv) * cpm)
            if role == ROLE_GYM_DEFENDER:
                kwargs["maxHP"] *= 2

        self.pokeType1 = kwargs["pokeType1"]
        self.pokeType2 = kwargs["pokeType2"]
        self.attack = kwargs["attack"]
        self.defense = kwargs["defense"]
        self.maxHP = kwargs["maxHP"]

        # Set up moves
        pvp = (role == ROLE_PVP_ATTACKER) or kwargs.get("pvp", False)
        if "fmove" in kwargs:
            self.fmove = Move(kwargs["fmove"], pvp=pvp,
                              game_master=game_master)
        raw_cmoves = []
        if "cmove" in kwargs:
            raw_cmoves = [kwargs["cmove"]]
            if "cmove2" in kwargs:
                raw_cmoves.append(kwargs["cmove2"])
        elif "cmoves" in kwargs:
            raw_cmoves = kwargs["cmoves"]
        cmoves = []
        for move in raw_cmoves:
            cmoves.append(Move(move, pvp=pvp))
        self.cmoves = cmoves

        # Set up other attributes
        self.immortal = kwargs.get("immortal", False)
        if "num_shields" in kwargs or "strategy2" in kwargs or "shield" in kwargs:
            self.num_shields = kwargs.get(
                "num_shields", kwargs.get("strategy2", kwargs.get("shield")))




'''
This module provide methods to access GoBattleSim engine.
'''

import argparse
from ctypes import *
import json
import os
import platform
import sys

if platform.system() == "Windows":
    lib = CDLL(os.path.join(os.path.dirname(__file__), "libGoBattleSim.dll"))
else:
    lib = CDLL(os.path.join(os.path.dirname(__file__), "libGoBattleSim.so"))


class GBS:

    lib.GBS_version.argtypes = []
    lib.GBS_version.restype = c_char_p
    @staticmethod
    def version():
        return lib.GBS_version().decode()

    lib.GBS_error.argtypes = []
    lib.GBS_error.restype = c_char_p
    @staticmethod
    def error():
        return lib.GBS_error().decode()

    lib.GBS_config.argtypes = [c_char_p]
    lib.GBS_config.restype = c_char_p
    @staticmethod
    def config(game_master=None):
        if game_master is not None:
            g_str = json.dumps(game_master)
            lib.GBS_config(g_str.encode())
        g_str = lib.GBS_config(None).decode()
        return json.loads(g_str)

    lib.GBS_prepare.argtypes = [c_char_p]
    lib.GBS_prepare.restype = c_void_p
    @staticmethod
    def prepare(sim_input):
        in_str = json.dumps(sim_input)
        lib.GBS_prepare(in_str.encode())

    lib.GBS_run.argtypes = []
    lib.GBS_run.restype = c_void_p
    @staticmethod
    def run():
        lib.GBS_run()

    lib.GBS_collect.argtypes = []
    lib.GBS_collect.restype = c_char_p
    @staticmethod
    def collect():
        out_str = lib.GBS_collect().decode()
        return json.loads(out_str)


def print_version():
    print("GoBattleSim Engine " + GBS.version())


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument("sim_input", nargs="?",
                        help="simulation input json")
    parser.add_argument("-v", "--version", action="store_true",
                        help="show GBS engine version and exit")
    parser.add_argument("-c", "--config", nargs="?", default=False,
                        help="print game master, or (with argument) set game master by path")
    parser.add_argument("-o", "--out", type=argparse.FileType('w'), default=sys.stdout,
                        help="file to save simulation output")
    args = parser.parse_args()

    if args.version:
        print_version()
        return 0

    if args.config is None:
        print(json.dumps(GBS.config(), indent=4))
        return 0

    if not args.sim_input:
        print_version()
        return 0

    if args.config:
        with open(args.config) as fd:
            j = json.load(fd)
            GBS.config(j)

    with open(args.sim_input) as fd:
        j = json.load(fd)
    try:
        GBS.prepare(j)
        GBS.run()
        sim_output = GBS.collect()
    except Exception as e:
        gbs_error = GBS.error()
        if gbs_error:
            print("GBS Engine error:", gbs_error)
        else:
            print(str(e))
        return -1

    json.dump(sim_output, args.out, indent=4)


if __name__ == "__main__":
    main()




'''
This module provides methods to create, load, run and save battle matrix.
'''

import argparse
import copy
import csv
import json
import math
import sys

from .GameMaster import GameMaster
from .Pokemon import Pokemon

try:
    from .Engine import GBS
except Exception as e:
    GBS = e


def set_stats(pkm, league, game_master=None):
    '''
    set the core stats (pokeType1, pokeType2, attack, defense, maxHP) for Pokemon @param pkm according to @param league

    @param pkm dict-like object containing field "name", or all of the core stats to avoid GameMaster look-up
    @param league one of {"great", "ultra", "master"}, or an int for target cp
    @param game_master GameMaster to search stats data for
    '''
    if game_master is None:
        game_master = GameMaster.CurrentInstance

    CoreStats = ["pokeType1", "pokeType2", "attack", "defense", "maxHP"]
    CoreBaseStats = ["pokeType1", "pokeType2", "baseAtk", "baseDef", "baseStm"]

    if all([stat in pkm for stat in CoreStats]):
        return pkm

    if all([stat in pkm for stat in CoreBaseStats]):
        pass
    else:
        game_master.apply()
        species = game_master.search_pokemon(pkm["name"])
        if species is None:
            return None
        for stat in CoreBaseStats:
            pkm[stat] = species[stat]

    if league == "master":
        pkm["cpm"] = game_master.CPMultipliers[-1]
        pkm["atkiv"] = 15
        pkm["defiv"] = 15
        pkm["stmiv"] = 15
    else:
        if league == "ultra":
            target_cp = 2500
        elif league == "great":
            target_cp = 1500
        elif type(league) is int:
            target_cp = league
        else:
            raise Exception("bad league {}".format(league))
        target_cp = 2500 if league == "ultra" else 1500
        pkm["cpm"], pkm["atkiv"], pkm["defiv"], pkm["stmiv"] = Pokemon.infer_cpm_and_IVs(
            pkm["baseAtk"], pkm["baseDef"], pkm["baseStm"], target_cp)

    pkm["attack"] = (pkm["baseAtk"] + pkm["atkiv"]) * pkm["cpm"]
    pkm["defense"] = (pkm["baseDef"] + pkm["defiv"]) * pkm["cpm"]
    pkm["maxHP"] = math.floor((pkm["baseStm"] + pkm["stmiv"]) * pkm["cpm"])

    return pkm


def set_moves(pkm, game_master=None):
    '''
    set the moves for Pokemon @param pkm
    '''
    if type(pkm.get("fmove")) is dict and all([type(move) is dict for move in pkm.get("cmoves", [0])]):
        return pkm

    if game_master is None:
        game_master = GameMaster.CurrentInstance

    fmove = game_master.search_pvp_fmove(pkm["fmove"])
    if fmove is None:
        return None
    pkm["fmove"] = fmove

    cmoves = pkm.get("cmoves", [])
    if "cmove" in pkm:
        cmoves.append(pkm["cmove"])
    if "cmove2" in pkm:
        cmoves.append(pkm["cmove2"])
    cmoves = set(cmoves)

    pkm["cmoves"] = []
    for cmove in cmoves:
        cmove = game_master.search_pvp_cmove(cmove)
        if cmove:
            pkm["cmoves"].append(cmove)

    return pkm


def load_pokemon(file, fmt="tsv"):
    '''
    load pokemon list from file @param file.

    @file file object
    @fmt format of Pokemon list file, one of {"tsv", "csv", "json"}
    @return a list of Pokemon with set stats and moves
    '''
    pkm_list = []
    if fmt == "tsv":
        reader = csv.DictReader(file, dialect="excel-tab")
        pkm_list = [pkm for pkm in reader]
    elif fmt == "csv":
        reader = csv.DictReader(file)
        pkm_list = [pkm for pkm in reader]
    elif fmt == "json":
        pkm_list = json.load(file)
    else:
        raise Exception("bad Pokemon list format {}".format(fmt))

    return pkm_list


def minimize_pokemon(pkm_list):
    CoreFields = ["name", "pokeType1", "pokeType2",
                  "attack", "defense", "maxHP", "fmove", "cmoves"]
    pkm_list_minimized = []
    for pkm in pkm_list:
        pkm_min = {k: pkm[k] for k in CoreFields}
        if "fmove" in pkm_min and "movetype" in pkm_min["fmove"]:
            pkm_min["fmove"].pop("movetype")
        for cmove in pkm_min.get("cmoves", []):
            if "movetype" in cmove:
                cmove.pop("movetype")
        pkm_list_minimized.append(pkm_min)
    return pkm_list_minimized


def save_pokemon(pkm_list, file, fmt="csv"):
    '''
    save pokemon list to file @param file.

    @param pkm_list a list of Pokemon with set stats and moves
    @file file object
    @fmt format of Pokemon list file, one of {"tsv", "csv", "json"}
    '''
    if not pkm_list:
        return
    if fmt == "tsv" or fmt == "csv":
        fields = list(pkm_list[0].keys())
        if "cmoves" in fields:
            fields.remove("cmoves")
            if "cmove" not in fields:
                fields.append("cmove")
            if "cmove2" not in fields and any([pkm.get("cmoves", [])[1:] for pkm in pkm_list]):
                fields.append("cmove2")
        if fmt == "tsv":
            writer = csv.DictWriter(file, fields, dialect="excel-tab")
        else:
            writer = csv.DictWriter(file, fields)
        writer.writeheader()
        for pkm in pkm_list:
            pkm_copy = copy.copy(pkm)
            if "cmoves" in pkm_copy:
                cmoves = pkm_copy.pop("cmoves")
                pkm_copy["cmove"] = cmoves[0].get("name")
                if cmoves[1:]:
                    pkm_copy["cmove2"] = cmoves[1].get("name")
            if type(pkm_copy.get("fmove")) is dict:
                pkm_copy["fmove"] = pkm_copy["fmove"].get("name")
            writer.writerow(pkm_copy)
    elif fmt == "json":
        json.dump(pkm_list, file, indent=4)
    else:
        raise Exception("bad format {}".format(fmt))


def do_run_matrix(row_pkm, col_pkm=[], shield=0):
    '''
    actually run the Battle Matrix.

    @param row_pkm list of Pokemon objects
    @param col_pkm list of Pokemon objects
    @param shield shield setting
    @return matrix as 2D list
    '''

    reqInput = {
        "battleMode": "battlematrix",
        "rowPokemon": row_pkm,
        "colPokemon": col_pkm,
        "avergeByShield": shield != 0
    }

    GBS.prepare(reqInput)
    GBS.run()
    return GBS.collect()


def load_and_set_pokemon(filepath, league="master", game_master=None):
    if game_master is None:
        game_master = GameMaster.CurrentInstance

    with open(filepath) as fd:
        pkm_list = load_pokemon(fd, filepath.split('.')[-1])
        pkm_list_filtered = []
        for pkm in pkm_list:
            if not set_stats(pkm, league, game_master):
                continue
            if not set_moves(pkm, game_master):
                continue
            pkm_list_filtered.append(pkm)
    return pkm_list_filtered


def run_matrix(row_pkm, col_pkm=None, shield=-1, league="master", game_master=None):
    '''
    create and run Battle Matrix.

    @param row_pkm path to a Pokemon list file
    @param col_pkm path to a Pokemon list file
    @param shield shield setting
    @return matrix as 2D list
    '''
    if game_master is None:
        game_master = GameMaster.CurrentInstance

    row_pkm = load_and_set_pokemon(row_pkm, league, game_master)
    if col_pkm is not None:
        col_pkm = load_and_set_pokemon(col_pkm, league, game_master)
    else:
        col_pkm = []

    return do_run_matrix(row_pkm, col_pkm, shield)


def save_matrix(matrix, file, fmt="csv"):
    '''
    save battle matrix @param matrix to file @file with format @param fmt
    '''
    if fmt == "tsv":
        writer = csv.writer(file, dialect="excel-tab")
        writer.writerows(matrix)
    elif fmt == "csv":
        writer = csv.writer(file)
        writer.writerows(matrix)
    elif fmt == "json":
        json.dump(matrix, file)
    else:
        raise Exception("bad format {}".format(fmt))


def main():
    parser = argparse.ArgumentParser()

    parser.add_argument("row_pokemon",
                        help="path to a file containing list of Pokemon")
    parser.add_argument("col_pokemon", nargs='?',
                        help="path to a file containing list of Pokemon. If unset, will be set to row Pokemon")
    parser.add_argument("-s", "--shield", type=int, default=0,
                        help="shield strategy setting. -1 for average")
    parser.add_argument("-l", "--league", choices=["great", "ultra", "master"],
                        help="PvP league to decide Pokemon stats. If not set, will use raw Pokemon input")
    parser.add_argument("-c", "--config", required=True,
                        help="path to GBS game master json")
    parser.add_argument("-p", "--pokemon", action="store_true",
                        help="only show the parsed row Pokemon list")
    parser.add_argument("-z", "--minimize", action="store_true",
                        help="for showing the Pokemon list, keeping only the necessary fields")
    parser.add_argument("-i", "--input", action="store_true",
                        help="only show the battle matrix input")
    parser.add_argument("-f", "--format", choices=["tsv", "csv", "json"], default="csv",
                        help="matrix output format")
    parser.add_argument("-o", "--out",
                        help="file to store output matrix")
    args = parser.parse_args()

    if args.out is None:
        args.out = sys.stdout
    else:
        args.out = open(args.out, "w", newline="")

    gm = GameMaster()
    with open(args.config, encoding="utf8") as fd:
        gm.from_json(json.load(fd))
    gm.apply()

    row_pkm = load_and_set_pokemon(args.row_pokemon, args.league)
    if args.minimize:
        row_pkm = minimize_pokemon(row_pkm)

    if args.pokemon:
        save_pokemon(row_pkm, args.out, args.format)
        return 0

    if args.col_pokemon is not None:
        col_pkm = load_and_set_pokemon(args.col_pokemon, args.league)
        if args.minimize:
            col_pkm = minimize_pokemon(col_pkm)
    else:
        col_pkm = []

    reqInput = {
        "battleMode": "battlematrix",
        "rowPokemon": row_pkm,
        "colPokemon": col_pkm,
        "avergeByShield": args.shield != 0
    }

    if args.input:
        json.dump(reqInput, args.out, indent=4)
        return 0

    if type(GBS) is Exception:
        raise GBS

    GBS.config(gm.to_json())
    GBS.prepare(reqInput)
    GBS.run()
    matrix = GBS.collect()

    save_matrix(matrix, args.out, args.format)


if __name__ == "__main__":
    main()




'''
This module provides game master file parsing (into GoBattleSim compatible format).
'''

import argparse
import copy
import json
import re
import sys


PoketypeList = ["normal", "fighting", "flying", "poison", "ground", "rock", "bug", "ghost",
                "steel", "fire", "water", "grass", "electric", "psychic", "ice", "dragon", "dark", "fairy"]

InversedPoketypeList = dict([(name, i)
                             for i, name in enumerate(PoketypeList)])


def rm_underscores(Str, Category):
    if Category == 'p':
        return '-'.join([s.lower() for s in Str.split('_')][2:])
    elif Category == "fast":
        return ' '.join([s.lower() for s in Str.split('_')[:-1]])
    elif Category == "charged":
        return ' '.join([s.lower() for s in Str.split('_')])
    elif Category == 't':
        return Str.split('_')[-1].lower()


class GameMaster:
    '''
    The class parses the official Game Master json and organize the data.
    '''

    CurrentInstance = None

    def __init__(self, file=None):
        '''
        @param file the filepath to the game master json file.
        '''
        self.Pokemon = []
        self.PvEMoves = []
        self.PvPMoves = []
        self.CPMultipliers = []
        self.WeatherSettings = {}
        self.FriendAttackBonusMultipliers = []
        self.TypeEffectiveness = {}
        self.PvEBattleSettings = {}
        self.PvPBattleSettings = {}
        # Note: for now, raid tiers data are hard coded because they are not in the game master file.
        self.RaidTierSettings = [
            {"name": "1", "cpm": 0.6, "maxHP": 600, "timelimit": 180000},
            {"name": "2", "cpm": 0.67, "maxHP": 1800, "timelimit": 180000},
            {"name": "3", "cpm": 0.7300000190734863,
                "maxHP": 3600, "timelimit": 180000},
            {"name": "4", "cpm": 0.7900000214576721,
                "maxHP": 9000, "timelimit": 180000},
            {"name": "5", "cpm": 0.7900000214576721,
                "maxHP": 15000, "timelimit": 300000},
            {"name": "6", "cpm": 0.7900000214576721,
                "maxHP": 18750, "timelimit": 3000000}
        ]

        if file is not None:
            self.parse(file)

    def clear(self):
        '''
        Clear all data.
        '''
        self.__init__()
        return self

    def parse(self, file):
        '''
        Load and process a game master json file @param file.
        @param path to game master json.
        '''
        with open(file) as fd:
            gmdata = json.load(fd)

        for template in gmdata["itemTemplates"]:
            tid = template['templateId']

            # Match Pokemon
            if re.fullmatch(r'V\d+_POKEMON_.+', tid):
                pokemon = {}
                pkmInfo = template["pokemonSettings"]
                pokemon['dex'] = int(tid.split('_')[0][1:])
                pokemon['name'] = rm_underscores(tid, 'p')
                pokemon['pokeType1'] = rm_underscores(
                    pkmInfo['type'], 't')
                pokemon['pokeType2'] = rm_underscores(
                    pkmInfo.get('type2', 'none'), 't')
                pokemon['baseAtk'] = pkmInfo["stats"]["baseAttack"]
                pokemon['baseDef'] = pkmInfo["stats"]["baseDefense"]
                pokemon['baseStm'] = pkmInfo["stats"]["baseStamina"]
                pokemon['fastMoves'] = [rm_underscores(
                    s, "fast") for s in pkmInfo.get('quickMoves', [])]
                pokemon['chargedMoves'] = [rm_underscores(
                    s, "charged") for s in pkmInfo.get('cinematicMoves', '')]
                evolution = [s.lower()
                             for s in pkmInfo.get('evolutionIds', [])]
                if any(evolution):
                    pokemon['evolution'] = evolution
                if 'rarity' in pkmInfo:
                    pokemon['rarity'] = pkmInfo['rarity']

                self.Pokemon.append(pokemon)

            # Match move, either Fast or Charged
            elif re.fullmatch(r'V\d+_MOVE_.+', tid):
                moveInfo = template['moveSettings']
                move = {}
                move['movetype'] = "fast" if tid.endswith(
                    '_FAST') else "charged"
                move['name'] = rm_underscores(
                    moveInfo["movementId"], move['movetype'])
                move['pokeType'] = rm_underscores(moveInfo["pokemonType"], 't')
                move['power'] = int(moveInfo.get("power", 0))
                move['duration'] = int(moveInfo["durationMs"])
                move['dws'] = int(moveInfo["damageWindowStartMs"])
                move['energy'] = int(moveInfo.get("energyDelta", 0))

                self.PvEMoves.append(move)

            # Match PvP Moves
            elif re.fullmatch(r'COMBAT_V\d+_MOVE_.+', tid):
                moveInfo = template['combatMove']
                move = {}
                move['movetype'] = "fast" if tid.endswith(
                    '_FAST') else "charged"
                move['name'] = rm_underscores(
                    moveInfo["uniqueId"], move['movetype'])
                move['pokeType'] = rm_underscores(moveInfo["type"], 't')
                move['power'] = int(moveInfo.get("power", 0))
                move['duration'] = int(moveInfo.get('durationTurns', 0)) + 1
                move['energy'] = int(moveInfo.get("energyDelta", 0))
                if "buffs" in moveInfo:
                    move['effect'] = {
                        "activation_chance": moveInfo["buffs"]["buffActivationChance"],
                        "self_attack_stage_delta": moveInfo["buffs"].get("attackerAttackStatStageChange", 0),
                        "self_defense_stage_delta": moveInfo["buffs"].get("attackerDefenseStatStageChange", 0),
                        "target_attack_stage_delta": moveInfo["buffs"].get("targetAttackStatStageChange", 0),
                        "target_defense_stage_delta": moveInfo["buffs"].get("targetDefenseStatStageChange", 0)
                    }

                self.PvPMoves.append(move)

            # Match CPM's
            elif tid == 'PLAYER_LEVEL_SETTINGS':
                for cpm in template["playerLevel"]["cpMultiplier"]:
                    if self.CPMultipliers:
                        # Half level
                        self.CPMultipliers.append(
                            ((cpm**2 + self.CPMultipliers[-1]**2)/2)**0.5)
                    self.CPMultipliers.append(cpm)

            # Match Pokemon Types
            elif re.fullmatch(r'POKEMON_TYPE_.+', tid):
                pokemonType = rm_underscores(tid, 't')
                self.TypeEffectiveness[pokemonType] = {}
                for idx, mtp in enumerate(template["typeEffective"]["attackScalar"]):
                    self.TypeEffectiveness[pokemonType][PoketypeList[idx]] = mtp

            # Match PvE Battle settings
            elif tid == 'BATTLE_SETTINGS':
                self.PvEBattleSettings = template["battleSettings"]

            # Match PvP Battle settings
            elif tid == 'COMBAT_SETTINGS':
                for name, value in template["combatSettings"].items():
                    self.PvPBattleSettings[name] = value

            # Match PvP Battle settings for buff stats
            elif tid == 'COMBAT_STAT_STAGE_SETTINGS':
                for name, value in template["combatStatStageSettings"].items():
                    self.PvPBattleSettings[name] = value

            # Match weather settings
            elif re.fullmatch(r'WEATHER_AFFINITY_.+', tid):
                wname = template["weatherAffinities"]["weatherCondition"]
                if wname == 'OVERCAST':
                    wname = 'CLOUDY'
                self.WeatherSettings[wname] = [rm_underscores(
                    s, 't') for s in template["weatherAffinities"]["pokemonType"]]
            elif tid == 'WEATHER_BONUS_SETTINGS':
                self.PvEBattleSettings['weatherAttackBonusMultiplier'] = template["weatherBonusSettings"]["attackBonusMultiplier"]

            # Match friend settings
            elif re.fullmatch(r'FRIENDSHIP_LEVEL_\d+', tid):
                multiplier = template["friendshipMilestoneSettings"]["attackBonusPercentage"]
                self.FriendAttackBonusMultipliers.append(
                    {"name": tid, "multiplier": multiplier})

        self.FriendAttackBonusMultipliers.sort(key=lambda x: x["multiplier"])

    def to_json(self):
        '''
        Export this instance in json.
        '''
        return self.__dict__

    def from_json(self, source):
        '''
        The reverse of to_json().
        @param source a dict-like object
        '''
        for name, value in source.items():
            if name in self.__dict__:
                self.__dict__[name] = value

    def apply(self):
        '''
        Pass the data to simulator engine and apply, and
        set GameMaster.CurrentInstance to this instance.
        '''
        GameMaster.CurrentInstance = self

    @staticmethod
    def _search(_universe, criteria, _all):
        if isinstance(criteria, str):
            def cbfn(x):
                return x['name'].strip().lower() == criteria.strip().lower()
        elif isinstance(criteria, int):
            def cbfn(x):
                return x.get('dex') == criteria
        else:
            cbfn = criteria
        results = []
        for entity in _universe:
            if cbfn(entity):
                if _all:
                    results.append(entity)
                else:
                    return entity
        if _all:
            return results
        else:
            return None

    def search_pokemon(self, criteria, _all=False):
        return GameMaster._search(self.Pokemon, criteria, _all)

    def search_pve_fmove(self, criteria, _all=False):
        return GameMaster._search(filter(lambda x: x.get('movetype') == "fast", self.PvEMoves), criteria, _all)

    def search_pve_cmove(self, criteria, _all=False):
        return GameMaster._search(filter(lambda x: x.get('movetype') == "charged", self.PvEMoves), criteria, _all)

    def search_pvp_fmove(self, criteria, _all=False):
        return GameMaster._search(filter(lambda x: x.get('movetype') == "fast", self.PvPMoves), criteria, _all)

    def search_pvp_cmove(self, criteria, _all=False):
        return GameMaster._search(filter(lambda x: x.get('movetype') == "charged", self.PvPMoves), criteria, _all)

    def search_cpm(self, level):
        idx = round(2 * float(level) - 2)
        return self.CPMultipliers[idx]

    def search_weather(self, weather_name):
        for i, wname in enumerate(sorted(self.WeatherSettings.keys())):
            if wname.lower() == weather_name.lower():
                return i
        return -1

    def search_friend(self, friendship):
        '''
        Return the friend attack bonus multiplier of friendship {friendship}.
        Example: "FRIENDSHIP_LEVEL_1", or 1, or "1", "Great", or "great"
        '''
        friendship = str(friendship).strip().lower()
        alter_names = ["none", "good", "great", "ultra", "best"]
        for i, friend_setting in enumerate(self.FriendAttackBonusMultipliers):
            if friendship == str(i):
                return friend_setting["multiplier"]
            elif friendship == alter_names[i]:
                return friend_setting["multiplier"]
        return 1

    def search_raid_tier(self, tier):
        '''
        Return {boss cpm, boss max_hp, time limit} of tier {tier}.
        {tier} is string, such as "3" (tier 3).
        '''

        tier = str(tier)
        for rt in self.RaidTierSettings:
            if rt["name"] == tier:
                return rt
        return None


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("infile", type=str,
                        help="path to official game master json")
    parser.add_argument("-m", "--mini", action="store_true",
                        help="leaving out Pokemon and Move data")
    parser.add_argument("-o", "--out", type=argparse.FileType('w'), default=sys.stdout,
                        help="parsing output")
    args = parser.parse_args()

    gm = GameMaster(args.infile)
    j = gm.to_json()
    if args.mini:
        j.pop("Pokemon")
        j.pop("PvPMoves")
        j.pop("PvEMoves")
    json.dump(j, args.out, indent=4)


if __name__ == "__main__":
    main()




import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="gobattlesim",
    version="0.8.1",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ymenghank/GoBattleSim-Python",
    
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Development Status :: 4 - Beta"
    ],

    packages=setuptools.find_packages(),
    package_data={'gobattlesim': ['libGoBattleSim.dll', 'libGoBattleSim.so']},
)




'''
This module provides PokeQuery, a scheme to search Pokemon/Move.
'''

import argparse
import copy
import csv
import itertools
import json
import sys

from .GameMaster import PoketypeList, GameMaster


POKE_QUERY_LOGICAL_OPERATORS = {
    ':': 0,
    ',': 0,
    ';': 0,
    '&': 1,
    '|': 1,
    '!': 2
}


def BasicPokeQuery(query_str, pkm=None, movetype="fast"):
    '''
    Create a basic PokeQuery from string @param query_str.

    @param pkm subject Pokemon. This parameter is needed if the entity to search is Move
    @param movetype used with searching Move
    @return a predicate/callback that accepts one parameter (the entity to be examined)
    '''

    query_str = str(query_str).lower().strip(" *")

    # Default predicate for empty query
    if query_str == "":
        if pkm is not None:
            pd1 = BasicPokeQuery("current", pkm, movetype)
            pd2 = BasicPokeQuery("legacy", pkm, movetype)
            pd3 = BasicPokeQuery("exclusive", pkm, movetype)
            return lambda x: (pd1(x) or pd2(x) or pd3(x))
        else:
            return lambda x: False

    # Match by dex. For Pokemon
    elif query_str.isdigit():
        dex = int(query_str)

        def predicate(entity):
            return entity.get('dex') == dex
    elif query_str[:3] == 'dex':
        num_part = query_str[3:]
        if '-' in num_part:
            min_dex, max_dex = [int(v.strip())
                                for v in num_part.split('-')][:2]
        else:
            min_dex = max_dex = int(num_part)

        def predicate(entity):
            return min_dex <= entity.get('dex') <= max_dex

    # Match by type. For Pokemon, Move
    elif query_str in PoketypeList or query_str == 'none':
        def predicate(entity):
            return query_str in [entity.get('pokeType'), entity.get('pokeType1'), entity.get('pokeType2')]

    # Match by rarity Legendary. For Pokemon
    elif query_str == "legendary":
        def predicate(entity):
            return entity.get("rarity") == "POKEMON_RARITY_LEGENDARY"

    # Match by rarity Mythical. For Pokemon
    elif query_str == "mythic" or query_str == "mythical":
        def predicate(entity):
            return entity.get("rarity") == "POKEMON_RARITY_MYTHIC"

    # Match by current move availability
    elif query_str == "current":
        movepool = pkm.get(movetype + "Moves", [])

        def predicate(entity):
            return entity['name'] in movepool
    elif query_str == "legacy":
        movepool = pkm.get(movetype + "Moves_legacy", [])

        def predicate(entity):
            return entity['name'] in movepool
    elif query_str == "exclusive":
        movepool = pkm.get(movetype + "Moves_exclusive", [])

        def predicate(entity):
            return entity['name'] in movepool

    # Default: Match by name. For Pokemon, Move
    else:
        def predicate(entity):
            return query_str in entity['name']

    return predicate


def PokeQuery(query_str, pkm=None, movetype="fast"):
    '''
    Create a PokeQuery from string @param query_str.
    Supports logical operators and parenthesis.

    @param pkm subject Pokemon. This parameter is needed if the entity to search is Move.
    @param movetype used with searching Move
    @return a callback/predicate that accepts one parameter (the entity to be examined).
    '''

    global POKE_QUERY_LOGICAL_OPERATORS
    OPS = POKE_QUERY_LOGICAL_OPERATORS

    # Tokenize the input query string
    tokens = []
    tk = ""
    for c in query_str:
        if c in OPS or c in ['(', ')']:
            tk = tk.strip()
            if tk:
                tokens.append(tk)
            tokens.append(c)
            tk = ""
        else:
            tk += c
    tk = tk.strip()
    if tk:
        tokens.append(tk)

    vstack = []
    opstack = []

    def default_pred(x):
        return False

    def eval_simple(op, vstack):
        if OPS[op] == 0:
            if len(vstack) < 2:
                raise Exception(f"missing operand for operator {op}")
            rhs = vstack.pop()
            lhs = vstack.pop()
            vstack.append(lambda x: lhs(x) or rhs(x))
        elif OPS[op] == 1:
            if len(vstack) < 2:
                raise Exception(f"missing operand for operator {op}")
            rhs = vstack.pop()
            lhs = vstack.pop()
            vstack.append(lambda x: lhs(x) and rhs(x))
        elif OPS[op] == 2:
            if len(vstack) < 1:
                raise Exception(f"missing operand for operator {op}")
            rhs = vstack.pop()
            vstack.append(lambda x: not rhs(x))

    for tk in tokens:
        if tk in OPS:
            while opstack and opstack[-1] != '(' and OPS[opstack[-1]] > OPS[tk]:
                eval_simple(opstack.pop(), vstack)
            opstack.append(tk)
        elif tk == '(':
            opstack.append(tk)
        elif tk == ')':
            while opstack:
                op = opstack.pop()
                if op == '(':
                    break
                eval_simple(op, vstack)
        else:
            vstack.append(BasicPokeQuery(tk, pkm=pkm, movetype=movetype))
    while opstack:
        eval_simple(opstack.pop(), vstack)

    return vstack.pop() if vstack else default_pred


def get_unique_pokemon(pkm_list):
    '''
    remove duplicates where {cmove, cmove2} are the same set.
    '''
    unique_pkm_list = []
    for pkm in pkm_list:
        unique = True
        for pkm2 in unique_pkm_list:
            if set([pkm['cmove'], pkm.get('cmove2')]) == set([pkm2['cmove'], pkm2.get('cmove2')]):
                unique = any([v != pkm2.get(k) for k, v in pkm.items()
                              if k != 'cmove' and k != 'cmove2'])
        if unique:
            unique_pkm_list.append(pkm)
    return unique_pkm_list


def batch_pokemon(pkm_qry, game_master: GameMaster):
    '''
    Return a list of Pokemon-Move combinations that match the query input @param pkm_qry in GameMaster @param game_master.

    @param pkm_qry is a dict, which may have the following fields:

        name
        fmove
        cmove
        [optional] cmove2

    All fields must be str, and can be PokeQuery.
    '''
    matches = []

    species_qry = pkm_qry["name"]
    fmove_qry = pkm_qry["fmove"]
    cmove_qry = pkm_qry["cmove"]
    cmove2_qry = pkm_qry.get("cmove2", "")

    species_matches = []
    fmove_matches = []
    cmove_matches = []
    cmove2_matches = []

    species_direct_match = game_master.search_pokemon(species_qry)
    if species_direct_match:
        species_matches.append(species_direct_match)
    else:
        species_matches = game_master.search_pokemon(
            PokeQuery(species_qry), True)

    for species in species_matches:
        cur_matches = []

        fmove_direct_match = game_master.search_pve_fmove(fmove_qry)
        if fmove_direct_match:
            fmove_matches.append(fmove_direct_match)
        else:
            fmove_matches = game_master.search_pve_fmove(
                PokeQuery(fmove_qry, species, "fast"), True)

        cmove_direct_match = game_master.search_pve_cmove(cmove_qry)
        if cmove_direct_match:
            cmove_matches.append(cmove_direct_match)
        else:
            cmove_matches = game_master.search_pve_cmove(
                PokeQuery(cmove_qry, species, "charged"), True)

        cmove2_direct_match = game_master.search_pve_cmove(cmove2_qry)
        if cmove2_direct_match:
            cmove2_matches.append(cmove2_direct_match)
        else:
            cmove2_matches = game_master.search_pve_cmove(
                PokeQuery(cmove2_qry, species, "charged"), True)

        for fmove, cmove in itertools.product(fmove_matches, cmove_matches):
            pkm = copy.copy(pkm_qry)
            for k in species:
                pkm[k] = species[k]
            pkm["fmove"] = fmove["name"]
            pkm["cmove"] = cmove["name"]
            if cmove2_matches:
                for cmove2 in cmove2_matches:
                    if cmove2["name"] != cmove["name"]:
                        pkm = copy.copy(pkm)
                        pkm["cmove2"] = cmove2["name"]
                        cur_matches.append(pkm)
            else:
                cur_matches.append(pkm)

        if cmove2_matches:
            cur_matches = get_unique_pokemon(cur_matches)

        matches.extend(cur_matches)

    return matches


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument("query", type=str, nargs='+',
                        help="species_query [, fmove_query] [, cmove_query] [, cmove2_query]")
    parser.add_argument("-c", "--config", required=True,
                        help="path to GBS game master json")
    parser.add_argument("-n", "--number", action="store_true",
                        help="only show the number of matches")
    parser.add_argument("-v", "--verbose", action="store_true",
                        help="print out each match in details")
    parser.add_argument("-f", "--format", choices=["tsv", "csv", "json"], default="csv",
                        help="format of output")
    parser.add_argument("-o", "--out",
                        help="file to store output")
    args = parser.parse_args()

    if args.out is None:
        args.out = sys.stdout
    else:
        args.out = open(args.out, mode="w", newline="")

    gm = GameMaster()
    with open(args.config, encoding="utf8") as fd:
        gm.from_json(json.load(fd))
    gm.apply()

    fields = []
    matches = []

    if len(args.query) == 1:
        fields = ["name"]
        matches = list(filter(PokeQuery(args.query[0]), gm.Pokemon))
    elif len(args.query) >= 3:
        fields = ["name", "fmove", "cmove"]
        pkm_qry = {
            "name": args.query[0],
            "fmove": args.query[1],
            "cmove": args.query[2]
        }
        if len(args.query) >= 4:
            fields.append("cmove2")
            pkm_qry["cmove2"] = args.query[3]
        matches = batch_pokemon(pkm_qry, gm)
    else:
        print("cannot query fast move but not primary charged move")
        return -2

    number = len(matches)
    if args.number:
        print(number)
        return 0
    if number == 0:
        return 0

    if args.verbose:
        for attr in matches[0].keys():
            if attr not in fields:
                fields.append(attr)
    else:
        matches = [{k: pkm[k] for k in fields} for pkm in matches]

    if args.format == "tsv":
        writer = csv.DictWriter(args.out, fields, dialect="excel-tab")
        writer.writeheader()
        writer.writerows(matches)
    elif args.format == "csv":
        writer = csv.DictWriter(args.out, fields)
        writer.writeheader()
        writer.writerows(matches)
    elif args.format == "json":
        json.dump(matches, args.out, indent=4)

    return 0


if __name__ == "__main__":
    exit(main())
